(declare (standard-bindings)
	 (extended-bindings)
	 (not safe)
	 (block))

(declare (not core))
(include "points.scm")
(declare (core))

(define number-of-refinements 10)

(define triangulations (time (Triangulation->refined-triangulations coarse-square-triangulation number-of-refinements)))
(define spaces         (time (triangulations->linear-finite-element-spaces triangulations)))
(define projectors     (time (linear-finite-element-spaces->projectors spaces)))
(define injectors      (time (projectors->injectors projectors)))

(define fes (vector-ref spaces number-of-refinements))

(define t (vector-ref triangulations number-of-refinements))

(time (Triangulation-consistent? t))

(define cg-data (CG-build-data t fes))

(define fine-t (time (refine t)))

(define fine-fes (time (Triangulation->Linear-finite-element-space fine-t)))

(define L2-inner-product-operator (time (->Linear-finite-element-operator fine-fes)))

(time (add-mass-terms! L2-inner-product-operator (lambda (p) 1.0)))

(define H1-inner-product-operator (time (->Linear-finite-element-operator fine-fes)))

(time (add-stiffness-terms! H1-inner-product-operator (lambda (p) 1.0)))

(time (add-mass-terms! H1-inner-product-operator (lambda (p) 1.0)))

(define coarse->fine-operator (time (Operator-adjoint (coarse<-fine fes fine-fes))))

(define symmetric-tests
  (list 
   ;; We start with problems that have zero error (if we solve the linear problem exactly),
   ;; even with 0 refinements.
   (instantiate Problem-data
		description: "a=1, c=1, u=1, Robin, gamma=x (error = 0 with exact linear algebra)"
		a: (lambda (p) 1.0)
		c: (lambda (p) 1.0)
		gamma: (lambda (p) (Point-x p))
		u: (lambda (p) 1.)
		f: (lambda (p) 1.)
		g: (lambda (p) (Point-x p)))
   (instantiate Problem-data
		description: "a=1, c=1, u=x, Neumann (error = 0 with exact linear algebra)"
		a: (lambda (p) 1.0)
		c: (lambda (p) 1.0)
		u: (lambda (p) (Point-x p))
		f: (lambda (p) (Point-x p))
		g: (lambda (p)
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (cond ((fl= x 0.) -1.)
			     ((fl= x 1.) 1.)
			     (else
			      0.)))))
   (instantiate Problem-data
		description: "a=1+xy, c=1, u=x+y, Neumann (error = 0 with exact linear algebra)"
		a: (lambda (p) (fl+ 1.0 (fl* (Point-x p) (Point-y p))))
		c: (lambda (p) 1.0)
		u: (lambda (p) (fl+ (Point-x p) (Point-y p)))
		f: (lambda (p) 0.)
		g: (lambda (p)
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (cond ((fl= x 0.) -1.)
			     ((fl= x 1.) (fl+ 1. y))
			     ((fl= y 0.) -1.)
			     ((fl= y 1.) (fl+ 1. x))
			     (else (error "shouldn't happen"))))))
   ;; u(x,y) = x^2 - y^2 (is harmonic), Neumann boundary conditions,
   ;; a=c=1, b = 0, f = u, g = 0, 2, or -2 depending on the boundary
   (instantiate Problem-data
		description: "a=1, c=1, u=x^2-y^2, Neumann"
		a: (interpolate-procedure->Linear-finite-element-vector
		    (lambda (p) 1.0)
		    fes)
		c: (lambda (p) 1.0)
		f: (lambda (p) 
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (fl- (fl* x x) (fl* y y))))
		g: (lambda (p)
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (cond ((fl= x 0.0) 0.0)
			     ((fl= x 1.0) 2.0)
			     ((fl= y 0.0) 0.0)
			     ((fl= y 1.0) -2.0)
			     (else (error "shouldn't happen")))))
		u: (lambda (p) 
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (fl- (fl* x x) (fl* y y)))))
   (let* ((pi (fl* 4.0 (flatan 1.0)))
	  (coeff (fl+ 1.0 (fl* pi pi))))
     (instantiate Problem-data
		  description: "a=1, c=1, u=\\cos(\\pi x), Neumann"
		  a: (lambda (p) 1.0)
		  c: (interpolate-procedure->Linear-finite-element-vector
		      (lambda (p) 1.0)
		      fes)
		  f: (lambda (p) 
 		       (let ((x (Point-x p)))
			 (fl* coeff (flcos (fl* pi x)))))
		  u: (lambda (p) 
		       (let ((x (Point-x p)))
			 (flcos (fl* pi x))))))
   (let* ((pi (fl* 4.0 (flatan 1.0)))
	  (coeff (fl+ 1.0 (fl* pi pi)))
	  (minus-pi (fl- pi)))
     (instantiate Problem-data
		  description: "a=1, c=1, u=\\sin(\\pi x)+\\sin(\\pi y), Neumann"
		  a: (lambda (p) 1.0)
		  c:  (lambda (p) 1.0)
		  f: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl* coeff (fl+ (flsin (fl* pi x)) (flsin (fl* pi y))))))
		  g: (lambda (p) minus-pi)
		  u: (lambda (p)  
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl+ (flsin (fl* pi x)) (flsin (fl* pi y)))))))
   (let ((pi (fl* 4.0 (flatan 1.0))))
     (instantiate Problem-data
		  description: "a=11.0+10.0*\\cos\\pi x, c=1, u=xy, Neumann"
		  a: (lambda (p) 
		       (let ((x (Point-x p)))
			 (fl+ 11.0 (fl* 10. (flcos (fl* pi x))))))
		  c: (lambda (p) 1.0)
		  f: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl+ (fl* x y)
			      (fl* 10.0 pi (flsin (fl* pi x)) y))))
		  g: (lambda (p)
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (cond ((fl= x 0.0) (fl- (fl* 21.0 y)))
			       ((fl= x 1.0) y)
			       ((fl= y 0.0) (fl* (fl+ 11.0 (fl* 10.0 (flcos (fl* pi x))))
						 (fl- x)))
			       ((fl= y 1.0) (fl* (fl+ 11.0 (fl* 10.0 (flcos (fl* pi x))))
						 x))
			       (else
				(error "shouldn't happen")))))
		  u: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl* x y)))))
   (instantiate Problem-data
		description: "a=1, c=1+xy, u=x^2-y^2, Neumann"
		a: 1.
		c: (lambda (p)  
		      (let ((x (Point-x p))
			    (y (Point-y p)))
			(fl+ 1.0 (fl* x y))))
		f: (lambda (p)  
		      (let ((x (Point-x p))
			    (y (Point-y p)))
			(fl* (fl+ 1.0 (fl* x y))
			     (fl- (fl* x x) (fl* y y)))))
		g: (lambda (p)
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (cond ((fl= x 0.0) 0.0)
			     ((fl= x 1.0) 2.0)
			     ((fl= y 0.0) 0.0)
			     ((fl= y 1.0) -2.0)
			     (else (error "shouldn't happen")))))
		u: (lambda (p)
		     (let ((x (Point-x p))
			   (y (Point-y p)))
		       (fl- (fl* x x) (fl* y y)))))
   (let* ((pi (fl* 4.0 (flatan 1.0))))
     (instantiate Problem-data
		  description: "a=1, c=2+\\cos(\\pi y), u=\\cos(\\pi x), Neumann"
		  a: (lambda (p) 1.0)
		  c: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl+ 2.0 (flcos (fl* pi y)))))
		  f: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl* (flcos (fl* pi x))
			      (fl+ (fl* pi pi)
				   (fl* (fl+ 2.0 (flcos (fl* pi y))))))))
		  u: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (flcos (fl* pi x))))))
   (let* ((pi (fl* 4.0 (flatan 1.0)))
	  (coeff (fl+ 1.0 (fl* pi pi)))
	  (minus-pi (fl- pi)))
     (instantiate Problem-data
		  description: "a=1, c=1, gamma=2+\\cos(\\pi x), u=\\sin(\\pi x)+\\sin(\\pi y), Robin"
		  a: (lambda (p) 1.0)
		  c:  (lambda (p) 1.0)
		  gamma: (lambda (p)  
			   (let ((x (Point-x p))
				 (y (Point-y p)))
			     (fl+ 2.0 (flcos (fl* pi x)))))
		  f: (lambda (p)  
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl* coeff (fl+ (flsin (fl* pi x)) (flsin (fl* pi y))))))
		  g: (lambda (p)
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl+ minus-pi
			      (fl* (fl+ 2.0 (flcos (fl* pi x)))
				   (fl+ (flsin (fl* pi x)) (flsin (fl* pi y)))))))
		  u: (lambda (p)  
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl+ (flsin (fl* pi x)) (flsin (fl* pi y)))))))
   (let ((pi (fl* 4.0 (flatan 1.0))))
     (instantiate Problem-data
		  description: "a=11.0+10.0*\\cos\\pi x, c=1+xy, gamma=2+\\cos(\\pi x), u=xy, Robin"
		  a: (lambda (p) 
		       (let ((x (Point-x p)))
			 (fl+ 11.0 (fl* 10. (flcos (fl* pi x))))))
		  c: (lambda (p) 1.0)
		  f: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl+ (fl* x y)
			      (fl* 10.0 pi (flsin (fl* pi x)) y))))
		  gamma: (lambda (p) 
			   (let ((x (Point-x p))
				 (y (Point-y p)))
			     (fl+ 2.0 (flcos (fl* pi x)))))
		  g: (lambda (p)
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl+ (fl* (fl* x y)
				   (fl+ 2.0 (flcos (fl* pi x))))
			      (cond ((fl= x 0.0) (fl- (fl* 21.0 y)))
				    ((fl= x 1.0) y)
				    ((fl= y 0.0) (fl* (fl+ 11.0 (fl* 10.0 (flcos (fl* pi x))))
						      (fl- x)))
				    ((fl= y 1.0) (fl* (fl+ 11.0 (fl* 10.0 (flcos (fl* pi x))))
						      x))
				    (else
				     (error "shouldn't happen"))))))
		  u: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl* x y)))))))

;; we've defined the following nonsymmetric tests, and tried to solve them by setting up
;; the normal equations in MG-add-operator-and-rhs-information! (i.e., calculating m-transpose
;; and multiplying both m and rhs by m-transpose), but it didn't work at all, except for the
;; first problem for some reason (and even then it didn't work well).  So support
;; for nonsymmetric problems in multigrid has been disabled.

(define nonsymmetric-tests
  (let ()
    (define flow-vector (make-Point 10.0 5.0))
    (list 
     ;; u(x,y) = x^2 - y^2 (is harmonic), Neumann boundary conditions,
     ;; a=c=1, b = (10 5), f = u, g = 0, 2, or -2 depending on the boundary
     (instantiate Problem-data
		  description: "a=1, b = (10 5), c=1, u=x^2-y^2, Neumann"
		  a: (lambda (p) 1.0)
		  b: (lambda (p) flow-vector)
		  c: (lambda (p) 1.0)
		  f: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl- (fl+ (fl* x x) (fl* 20.0 x)) (fl+ (fl* y y) (fl* 10.0 y)))))
		  g: (lambda (p)
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (cond ((fl= x 0.0) 0.0)
			       ((fl= x 1.0) 2.0)
			       ((fl= y 0.0) 0.0)
			       ((fl= y 1.0) -2.0)
			       (else (error "shouldn't happen")))))
		  u: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl- (fl* x x) (fl* y y)))))
     (let* ((pi (fl* 4.0 (flatan 1.0)))
	    (coeff (fl+ 1.0 (fl* pi pi))))
       (instantiate Problem-data
		    description: "a=1, b = (10 5), c=1, u=\\cos(\\pi x), Neumann"
		    a: (lambda (p) 1.0)
		    b: (lambda (p) flow-vector)
		    c: (lambda (p) 1.0)
		    f: (lambda (p) 
			 (let ((x (Point-x p)))
			   (fl- (fl* coeff (flcos (fl* pi x))) (fl* 10.0 pi (flsin (fl* pi x))))))
		    u: (lambda (p) 
			 (let ((x (Point-x p)))
			   (flcos (fl* pi x))))))
     (let* ((pi (fl* 4.0 (flatan 1.0)))
	    (coeff (fl+ 1.0 (fl* pi pi)))
	    (minus-pi (fl- pi)))
       (instantiate Problem-data
		    description: "a=1, b = (10 5), c=1, u=\\sin(\\pi x)+\\sin(\\pi y), Neumann"
		    a: (lambda (p) 1.0)
		    b: (lambda (p) flow-vector)
		    c:  (lambda (p) 1.0)
		    f: (lambda (p) 
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl+ (fl* coeff (fl+ (flsin (fl* pi x)) (flsin (fl* pi y))))
				(fl* 10.0 pi (flcos (fl* pi x)))
				(fl* 5.0 pi (flcos (fl* pi y))))))
		    g: (lambda (p) minus-pi)
		    u: (lambda (p)  
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl+ (flsin (fl* pi x)) (flsin (fl* pi y)))))))
     (let ((pi (fl* 4.0 (flatan 1.0))))
       (instantiate Problem-data
		    description: "a=11.0+10.0*\\cos\\pi x, b = (10 5), c=1, u=xy, Neumann"
		    a: (lambda (p)  
			 (let ((x (Point-x p)))
			   (fl+ 11.0 (fl* 10. (flcos (fl* pi x))))))
		    b: (lambda (p) flow-vector)
		    c: (lambda (p) 1.0)
		    f: (lambda (p) 
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl+ (fl* x y)
				(fl* 10.0 pi (flsin (fl* pi x)) y)
				(fl* 10.0 y)
				(fl* 5.0 x))))
		    g: (lambda (p)
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (cond ((fl= x 0.0) (fl- (fl* 21.0 y)))
				 ((fl= x 1.0) y)
				 ((fl= y 0.0) (fl* (fl+ 11.0 (fl* 10.0 (flcos (fl* pi x))))
						   (fl- x)))
				 ((fl= y 1.0) (fl* (fl+ 11.0 (fl* 10.0 (flcos (fl* pi x))))
						   x))
				 (else
				  (error "shouldn't happen")))))
		    u: (lambda (p) 
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl* x y)))))
     (instantiate Problem-data
		  description: "a=1, b = (10 5), c=1+xy, u=x^2-y^2, Neumann"
		  a: (lambda (p) 1.0)
		  b: (lambda (p) flow-vector)
		  c: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl+ 1.0 (fl* x y))))
		  f: (lambda (p) 
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl+ (fl* (fl+ 1.0 (fl* x y))
				   (fl- (fl* x x) (fl* y y)))
			      (fl- (fl* 20.0 x)
				   (fl* 10.0 y)))))
		  g: (lambda (p)
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (cond ((fl= x 0.0) 0.0)
			       ((fl= x 1.0) 2.0)
			       ((fl= y 0.0) 0.0)
			       ((fl= y 1.0) -2.0)
			       (else (error "shouldn't happen")))))
		  u: (lambda (p)
		       (let ((x (Point-x p))
			     (y (Point-y p)))
			 (fl- (fl* x x) (fl* y y)))))
     (let* ((pi (fl* 4.0 (flatan 1.0))))
       (instantiate Problem-data
		    description: "a=1, b = (10 5), c=2+\\cos(\\pi y), u=\\cos(\\pi x), Neumann"
		    a: (lambda (p) 1.0)
		    b: (lambda (p) flow-vector)
		    c: (lambda (p) 
			 (let ((y (Point-y p)))
			   (fl+ 2.0 (flcos (fl* pi y)))))
		    f: (lambda (p) 
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl- (fl* (flcos (fl* pi x))
				     (fl+ (fl* pi pi)
					  (fl* (fl+ 2.0 (flcos (fl* pi y))))))
				(fl* 10.0 pi (flsin (fl* pi x))))))
		    u: (lambda (p) 
			 (let ((x (Point-x p)))
			   (flcos (fl* pi x))))))
     (let* ((pi (fl* 4.0 (flatan 1.0)))
	    (coeff (fl+ 1.0 (fl* pi pi)))
	    (minus-pi (fl- pi)))
       (instantiate Problem-data
		    description: "a=1, b = (10 5), c=1, gamma=2+\\cos(\\pi x), u=\\sin(\\pi x)+\\sin(\\pi y), Robin"
		    a: (lambda (p) 1.0)
		    b: (lambda (p) flow-vector)
		    c:  (lambda (p) 1.0)
		    gamma: (lambda (p)  
			     (let ((x (Point-x p)))
			       (fl+ 2.0 (flcos (fl* pi x)))))
		    f: (lambda (p) 
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl+ (fl* coeff (fl+ (flsin (fl* pi x)) (flsin (fl* pi y))))
				(fl* 10.0 pi (flcos (fl* pi x)))
				(fl* 5.0 pi (flcos (fl* pi y))))))
		    g: (lambda (p)
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl+ minus-pi
				(fl* (fl+ 2.0 (flcos (fl* pi x)))
				     (fl+ (flsin (fl* pi x)) (flsin (fl* pi y)))))))
		    u: (lambda (p) 
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl+ (flsin (fl* pi x)) (flsin (fl* pi y)))))))
     (let ((pi (fl* 4.0 (flatan 1.0))))
       (instantiate Problem-data
		    description: "a=11.0+10.0*\\cos\\pi x, b = (10 5), c=1+xy, gamma=2+\\cos(\\pi x), u=xy, Robin"
		    a: (lambda (p) 
			 (let ((x (Point-x p)))
			   (fl+ 11.0 (fl* 10. (flcos (fl* pi x))))))
		    b: (lambda (p) flow-vector)
		    c: (lambda (p) 1.0)
		    f: (lambda (p)
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl+ (fl* x y)
				(fl* 10.0 pi (flsin (fl* pi x)) y)
				(fl* 10.0 y)
				(fl* 5.0 x))))
		    gamma: (lambda (p) 
			     (let ((x (Point-x p)))
			       (fl+ 2.0 (flcos (fl* pi x)))))
		    g: (lambda (p)
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl+ (fl* (fl* x y)
				     (fl+ 2.0 (flcos (fl* pi x))))
				(cond ((fl= x 0.0) (fl- (fl* 21.0 y)))
				      ((fl= x 1.0) y)
				      ((fl= y 0.0) (fl* (fl+ 11.0 (fl* 10.0 (flcos (fl* pi x))))
							(fl- x)))
				      ((fl= y 1.0) (fl* (fl+ 11.0 (fl* 10.0 (flcos (fl* pi x))))
							x))
				      (else
				       (error "shouldn't happen"))))))
		    u: (lambda (p) 
			 (let ((x (Point-x p))
			       (y (Point-y p)))
			   (fl* x y))))))))


(define (run-test check solver-thunk)
  
  (define (format-for-display x)
    (if (fl< x 1.0)
	(let ((x (fl+ 1.0 x)))
	  (let ((string-x (number->string x)))
	    (string-append (substring string-x 1 (fxmin 12 (string-length string-x)))
			   (make-string (fx- 12 (fxmin 12 (string-length string-x))) #\0))))
	(number->string x)))
  
  (let* ((at-start (##process-statistics))
	 (result (solver-thunk))
	 (at-end (##process-statistics))
	 (user+sys-ns (fl+ (fl- (f64vector-ref at-end 0) (f64vector-ref at-start 0))
			   (fl- (f64vector-ref at-end 1) (f64vector-ref at-start 1))))
	 (gc-ns (fl+ (fl- (f64vector-ref at-end 3) (f64vector-ref at-start 3))
		     (fl- (f64vector-ref at-end 4) (f64vector-ref at-start 4))))
	 (difference (Vector-subtract (Operator-apply coarse->fine-operator result) check)))
    
    (for-each display (list #\tab
			    (format-for-display (sqrt (Vector-inner-product difference L2-inner-product-operator difference)))))
    (for-each display (list #\tab
			    (format-for-display (sqrt (Vector-inner-product difference H1-inner-product-operator difference)))))
    (for-each display (list #\tab
			    (round (fl* 1000. (fl- user+sys-ns gc-ns)))))
    (for-each display (list #\tab
			    (round (fl* 1000. gc-ns))
			    ))
    (newline)))

(define (run-multigrid-tests tests)
  
  (for-each (lambda (problem-data)
	      (with-access
	       problem-data (Problem-data a c f gamma g u)
	       (let* ((check            (interpolate-procedure->Linear-finite-element-vector u fine-fes))
		      (operator         (build-operator a c gamma fes))
		      (rhs              (build-rhs f g fes))
		      (right-hand-sides (most-refined-rhs->all-refined-rhs rhs projectors))
		      (operators        (most-refined-operator->all-refined-operators operator projectors injectors)))
		 
		 (for-each display (list #\newline (Problem-data-description problem-data) #\newline))
		 
		 (for-each (lambda (smoother-maker smoother-name)
			     (let ((smoothers! (make-smoothers operators smoother-maker)))
			       (for-each (lambda (p)
					   (for-each (lambda (m1)
						       
						       (define (iota m)
							 (let loop ((result '()) (m m))
							   (if (fx< m 0)
							       result
							       (loop (cons m result) (fx- m 1)))))
						       
						       (for-each (lambda (m2)
								   (for-each (lambda (r)
									       (let ((mg-data (instantiate Multigrid-data
													   triangulations:   triangulations
													   spaces:           spaces
													   projectors:       projectors
													   injectors:        injectors
													   operators:        operators
													   smoothers!:       smoothers!)))
										 
										 (MG-add-solver-information! mg-data smoother-name r m1 m2 p)
										 (display (Multigrid-data-solver-description mg-data))
										 (run-test check (lambda () (MG-solver mg-data right-hand-sides)))))
									     '(1 2 3 4)))
								 (list m1)));(iota m1)))
						     '(1 2 3 4)))
					 '(1 2))))
			   (list richardson-smoother!
				 jacobi-smoother!
				 gauss-seidel-smoother!)
			   (list "  Richardson"
				 "      Jacobi"
				 "Gauss-Seidel")))))
	    tests))

(define (run-cg-tests tests)
  
  (for-each display (list "The number of refinements is " number-of-refinements #\newline))
  
  (for-each (lambda (problem-data)

	      (for-each display (list #\newline (Problem-data-description problem-data) #\newline))
	      
	      (CG-add-operator-and-rhs-information! cg-data problem-data)
	      
	      (for-each (lambda (preconditioner preconditioner-name)
			  
			  (define check (interpolate-procedure->Linear-finite-element-vector (Problem-data-u problem-data) fine-fes))
			  
			  (CG-add-solver-information! cg-data
						      (preconditioner (Conjugate-gradient-data-operator cg-data))
						      preconditioner-name 
						      2000
						      (expt (Vector-dimension (Conjugate-gradient-data-right-hand-side cg-data)) -2.0))
			  (display (Conjugate-gradient-data-solver-description cg-data))
			  (run-test check (Conjugate-gradient-data-solver cg-data)))
			
			(list (lambda (m) identity-operator)
			      diagonal-preconditioner
			      gauss-seidel-preconditioner)
			'("(Identity preconditioner)    "
			  "(Diagonal preconditioner)    "
			  "(Gauss-Seidel preconditioner)")))
	    tests))
