;; The multigrid algorithm for solving linear systems requires a lot of structure---
;; a family of nested (in our case) triangulations, the associated finite-element spaces,
;; projectors on the duals of the spaces, injectors on the spaces themselves, the finite-element
;; operator and right-hand-side at each level to be solved, and a bunch of solvers.

;; Once you have that ;-), the algorithm is trivial.

(define-class Multigrid-data Object
  ((= triangulations maybe-uninitialized:)
   (= spaces maybe-uninitialized:)

   ;; to project on coarser grid use (vector-ref projectors k) v

   (= projectors maybe-uninitialized:)

   ;; to inject from a coarser grid, use (vector-ref injectors k) v

   (= injectors maybe-uninitialized:)

   ;; we're trying to solve (vector-ref operators k) z = (vector-ref right-hand-sides k)
   ;; The right-hand-sides vector will need to be calculated in MG-solve

   (= operators maybe-uninitialized:)

   ;; a vector of smoothers for each level
   (= smoothers! maybe-uninitialized:)

   ;; a vector of iterative solvers to be used at each level

   (= solvers maybe-uninitialized:)
   (= solver-description maybe-uninitialized:)))


(define-generic (MG-solver (data) right-hand-sides))

(define-handy-method (MG-solver (data Multigrid-data) right-hand-sides)
  (let* ((n (vector-length operators))
	 (results (make-vector n)))
    (vector-set! results 0 ((vector-ref solvers 0) (vector-ref right-hand-sides 0)))
    (let main-loop ((k 1))
      (if (fx< k n)
	  (begin
	    
	    ;; Solve at each level, using the injection of the solution at the previous level
	    ;; as the initial guess unless k is zero, where we use zero as the initial guess.
	    
	    (vector-set! results k ((vector-ref solvers k)
				    (vector-ref right-hand-sides k)
				    (Operator-apply (vector-ref injectors k)
						    (vector-ref results (fx- k 1)))))
	    (main-loop (fx+ k 1)))
	  (vector-ref results (fx- n 1))))))

(define (Triangulation->refined-triangulations coarse-triangulation number-of-refinements)
  (let ((result (make-vector (fx+ number-of-refinements 1))))
    (vector-set! result 0 coarse-triangulation)
    (do ((i 1 (fx+ i 1)))
	((fx> i number-of-refinements) result)
      (vector-set! result i (refine (vector-ref result (fx- i 1)))))))

(define (triangulations->linear-finite-element-spaces triangulations)
  (let* ((n (vector-length triangulations))
	 (result (make-vector n)))
    (do ((i 0 (fx+ i 1)))
	((fx= i n) result)
      (vector-set! result i (Triangulation->Linear-finite-element-space (vector-ref triangulations i))))))

(define (linear-finite-element-spaces->projectors spaces)
  (let* ((n (vector-length spaces))
	 (result (make-vector n)))
    (do ((i 1 (fx+ i 1)))
	((fx= i n) result)
      (vector-set! result i (coarse<-fine (vector-ref spaces (fx- i 1)) 
					  (vector-ref spaces i))))))

(define (projectors->injectors projectors)
  (let* ((n (vector-length projectors))
	 (result (make-vector n)))
    (do ((i 1 (fx+ i 1)))
	((fx= i n) result)
      (vector-set! result i (projector->injector (vector-ref projectors i))))))

(define (most-refined-operator->all-refined-operators operator projectors injectors)
  (let* ((n (vector-length projectors))
	 (result (make-vector n)))
    (vector-set! result (fx- n 1) operator)
    (do ((i (fx- n 1) (fx- i 1)))
	((fx= i 0) result)
      (vector-set! result (fx- i 1) (Operator-compose (vector-ref projectors i)
						      (Operator-compose (vector-ref result i)
									(vector-ref injectors i)))))))

(define (most-refined-rhs->all-refined-rhs rhs projectors)
  (let* ((n (vector-length projectors))
	 (result (make-vector n)))
    (vector-set! result (fx- n 1) rhs)
    (do ((i (fx- n 1) (fx- i 1)))
	((fx= i 0) result)
      (vector-set! result (fx- i 1) (Operator-apply (vector-ref projectors i)
						    (vector-ref result i))))))

(define (build-operator a c gamma fes)
  ;; gamma may be #f, meaning no gamma
  ;; fes is a linear-finite-element-space, usually the most refined one
  (let ((m (->Linear-finite-element-operator fes)))
    (add-stiffness-terms! m a)
    (add-mass-terms!      m c)
    (if gamma
	(add-boundary-terms! m gamma))
    m))

(define (build-rhs f g fes)
  (let ((rhs (instantiate Linear-finite-element-vector
			  space: fes)))
    (add-interior-integral-terms! rhs f)
    (if g
	(add-boundary-integral-terms! rhs g))
    rhs))


(define (make-smoothers operators smoother-maker)
  (let* ((n (vector-length operators))
	 (result (make-vector n)))
    (do ((i 0 (fx+ i 1)))
	((fx= i n) result)
      (vector-set! result i (smoother-maker (vector-ref operators i))))))

;;; Single multigrid step.  See Brenner and Scott for notation.

(define (MG-step
	 data                       ; Multigrid-data
	 m1                         ; number of pre-smoothings
	 m2                         ; number of post-smoothings
	 p                          ; number of coarse corrections
	 k                          ; current refinement level, 0 is coarsest grid, k > 0
	 z0                         ; initial guess
	 rhs                        ; right-hand-side
	 )

  (with-access
   data (Multigrid-data operators projectors injectors solvers smoothers!)
   (if (fx= k 0)
       ((vector-ref solvers 0) rhs)
       (let ((smoother! (vector-ref smoothers! k)))
	 
	 (let pre-smoothing-loop ((zl z0) (l 0))
	   (if (fx< l m1)
	       (pre-smoothing-loop (smoother! rhs zl)
				   (fx+ l 1))
	       
	       (let ((post-coarse-correction
		      (let ((coarse-rhs   (Operator-apply (vector-ref projectors k) 
							  (Vector-subtract rhs (Operator-apply (vector-ref operators k) zl)))))
			(let correction-loop ((q (Vector-zero coarse-rhs)) (l 0))
			  (if (fx< l p)
			      (correction-loop 
			       (MG-step data m1 m2 p (fx- k 1) q coarse-rhs)
			       (fx+ l 1))
			      (Vector-add zl (Operator-apply (vector-ref injectors k) q)))))))
		 
		 (let post-smoothing-loop ((zl post-coarse-correction) (l 0))
		   (if (fx< l m2)
		       (post-smoothing-loop (smoother! rhs zl)
					    (fx+ l 1))
		       zl)))))))))



(define-generic (MG-add-solver-information! (multigrid-data) smoother-name r m1 m2 p))

(define-handy-method (MG-add-solver-information! (multigrid-data Multigrid-data) smoother-name r m1 m2 p)
  (set! solver-description (string-append "MG: "
					  smoother-name
					  ", r = "
					  (number->string r)
					  ", m1 = "
					  (number->string m1)
					  ", m2 = "
					  (number->string m2)
					  ", p = "
					  (number->string p)
					  ".  "))
  (let ((n (vector-length operators)))
    (set! solvers (make-vector n))

    (vector-set! solvers 0 (let* ((m
				     (->Sparse-matrix
				      (Linear-finite-element-operator-matrix-implementation 
				       (vector-ref operators 0))))
				    (U
				     (Sparse-matrix-Cholesky-factor m)))
			     (lambda (b)
			       (instantiate-from (b Finite-element-vector)
						 data: (Sparse-matrix-solve-Ux=b
							U (Sparse-matrix-solve-U^Tx=b
							   U (Finite-element-vector-data b)))))))
    
    (do ((k 1 (fx+ k 1)))
	((fx= k n))
      (vector-set! solvers k
		   (lambda (b initial-guess)
		     (do ((i 0 (fx+ i 1))
			  (z initial-guess (MG-step multigrid-data
						    m1
						    m2
						    p
						    k 
						    z
						    b)))
			 ((fx= i r) z)))))))

